@@ -78,6 +78,7 @@ void    smtpd_state_init(SMTPD_STATE *state, VSTREAM *stream,
      * Initialize the state information for this connection, and fill in the
      * connection-specific fields.
      */
+    state->flags = 0;
     state->err = CLEANUP_STAT_OK;
     state->client = stream;
     state->service = mystrdup(service);
@@ -122,6 +123,7 @@ void    smtpd_state_init(SMTPD_STATE *state, VSTREAM *stream,
     state->proxy_xforward_features = 0;
     state->saved_filter = 0;
     state->saved_redirect = 0;
+    state->saved_bcc = 0;
     state->saved_flags = 0;
 #ifdef DELAY_ACTION
     state->saved_delay = 0;
@@ -136,6 +138,7 @@ void    smtpd_state_init(SMTPD_STATE *state, VSTREAM *stream,
 #endif
     state->dsn_envid = 0;
     state->dsn_buf = vstring_alloc(100);
+    state->dsn_orcpt_buf = vstring_alloc(100);
 #ifdef USE_TLS
     state->tls_use_tls = 0;
     state->tls_enforce_tls = 0;
@@ -150,6 +153,9 @@ void    smtpd_state_init(SMTPD_STATE *state, VSTREAM *stream,
 	smtpd_sasl_connect(state, VAR_SMTPD_SASL_OPTS, var_smtpd_sasl_opts);
 #endif
 
+    state->milter_argv = 0;
+    state->milter_argc = 0;
+
     /*
      * Initialize peer information.
      */
@@ -182,6 +188,8 @@ void    smtpd_state_reset(SMTPD_STATE *state)
 	vstring_free(state->buffer);
     if (state->addr_buf)
 	vstring_free(state->addr_buf);
+    if (state->access_denied)
+	myfree(state->access_denied);
     if (state->protocol)
 	myfree(state->protocol);
     smtpd_peer_reset(state);
@@ -206,6 +214,8 @@ void    smtpd_state_reset(SMTPD_STATE *state)
 	vstring_free(state->instance);
     if (state->dsn_buf)
 	vstring_free(state->dsn_buf);
+    if (state->dsn_orcpt_buf)
+	vstring_free(state->dsn_orcpt_buf);
 
 #ifdef USE_SASL_AUTH
     if (var_smtpd_sasl_enable)
